@@ -2,6 +2,8 @@
 
 set -eu -o pipefail
 
+pip install twine sphinx
+
 changelog=$(cat HISTORY.rst)
 
 regex='
@@ -21,7 +23,7 @@ version="${BASH_REMATCH[1]}"
 date="${BASH_REMATCH[2]}"
 notes="$(echo "${BASH_REMATCH[3]}" | sed -n -e '/^[0-9]\+\.[0-9]\+\.[0-9]\+/,$!p')"
 
-if [[ "$date" -ne  $(date +"%Y-%m-%d") ]]; then
+if [[ "$date" !=  $(date +"%Y-%m-%d") ]]; then
     echo "$date is not today!"
     exit 1
 fi
@@ -33,7 +35,7 @@ if [ -n "$(git status --porcelain)" ]; then
     exit 1
 fi
 
-perl -pi -e "s/(?<=__version__ = ').+?(?=')/$version/gsm" maxminddb/__init__.py
+perl -pi -e "s/(?<=__version__ = \").+?(?=\")/$version/gsm" maxminddb/__init__.py
 
 echo $"Test results:"
 python setup.py test
@@ -63,4 +65,6 @@ hub release create -m "$message" "$tag"
 
 git push --tags
 
-python setup.py release
+rm -fr dist
+python setup.py build_html sdist
+twine upload dist/*
